/************************************************************************/
/*                                                                      */
/*                      RC5 Remote Receiver                             */
/*                                                                      */
/*              Author: Peter Dannegger                                 */
/*                      danni@specs.de                                  */
/*                                                                      */
/************************************************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>

#include "rc5.h"


#define RC5TIME 	1.778e-3		// 1.778msec
#define PULSE_MIN	(uchar)(XTAL / 512 * RC5TIME * 0.4 + 0.5)
#define PULSE_1_2	(uchar)(XTAL / 512 * RC5TIME * 0.8 + 0.5)
#define PULSE_MAX	(uchar)(XTAL / 512 * RC5TIME * 1.2 + 0.5)


uchar	rc5_bit;				// bit value
uchar	rc5_time;				// count bit time
uint	rc5_tmp;				// shift bits in
volatile uint	rc5_data;				// store result

/////////////////////////////////////////////////////////
/// User functions 
/////////////////////////////////////////////////////////
void init_rc5()
{
   TCCR0B = 1<<CS02;			//divide by 256
   TIMSK = 1<<TOIE0;			//enable timer interrupt
}

uchar read_rc5(uint8_t* addr, uint8_t* code)
{
   uint i;
   cli();
   i = rc5_data;			// read two bytes from interrupt !
   rc5_data = 0;
   sei();
   
   static uint8_t old_tog = 0;

   uint8_t tmp;

   if(i) {
      uint8_t tog = (i >> 11) & 0x1;
      if(tog == old_tog)
	    return 0;
      old_tog = tog;

      tmp = (i >> 6) & 0x1F;
      if (addr) *addr = tmp;

      tmp = (i & 0x3F) | (~i >> 7 &0x40);
      if (code) *code = tmp;
      return 1;
      
   }
   else
   {
      if(code)
	 *code = 0;
      if(addr)
	 *addr = 0;
      return 0;
      
   }
}


////////////////////////////////////////////////////////////////////
// Internal functions 
////////////////////////////////////////////////////////////////////
SIGNAL (SIG_OVERFLOW0)
{
  uint tmp = rc5_tmp;				// for faster access

  TCNT0 = -2;					// 2 * 256 = 512 cycle

  if( ++rc5_time > PULSE_MAX ){			// count pulse time
    if( !(tmp & 0x4000) && tmp & 0x2000 )	// only if 14 bits received
      rc5_data = tmp;
    //lcd_clear();
	tmp = 0;
  }

  //lcd_home();
  //char s[30];
  //itoa(rc5_time, s, 10);
 // lcd_string(s);

  if( (rc5_bit ^ xRC5_IN) & 1<<xRC5 ){		// change detect
    rc5_bit = ~rc5_bit;				// 0x00 -> 0xFF -> 0x00
	//lcd_data('.');
    if( rc5_time < PULSE_MIN )			// to short
    {
	  //lcd_setcursor(0, 2);
	  //lcd_data('!');
	  tmp = 0;
	}

    if( !tmp || rc5_time > PULSE_1_2 ){		// start or long pulse time
      if( !(tmp & 0x4000) )			// not to many bits
        tmp <<= 1;				// shift
      if( !(rc5_bit & 1<<xRC5) )		// inverted bit
        tmp |= 1;				// insert new bit
      rc5_time = 0;				// count next pulse time
	  //char s[10];
	  //itoa(tmp, s, 10);
	  
	  //lcd_home();
	  //lcd_string(s);
    }
  }

  rc5_tmp = tmp;
}
